@@ -85,19 +85,24 @@ struct browse_req {
 	DBusConnection *conn;
 	DBusMessage *msg;
 	struct btd_device *device;
+	GSList *uuids;
 	GSList *uuids_added;
 	GSList *uuids_removed;
 	sdp_list_t *records;
 	int search_uuid;
-	gboolean browse;
 };
 
 static uint16_t uuid_list[] = {
 	PUBLIC_BROWSE_GROUP,
+	PNP_INFO_SVCLASS_ID,
 	HID_SVCLASS_ID,
 	GENERIC_AUDIO_SVCLASS_ID,
 	ADVANCED_AUDIO_SVCLASS_ID,
 	AV_REMOTE_SVCLASS_ID,
+	PANU_SVCLASS_ID,
+	GN_SVCLASS_ID,
+	NAP_SVCLASS_ID,
+	SERIAL_PORT_SVCLASS_ID,
 	0
 };
 
@@ -106,10 +111,15 @@ static GSList *device_drivers = NULL;
 static void device_free(gpointer user_data)
 {
 	struct btd_device *device = user_data;
+	struct btd_adapter *adapter = device->adapter;
+	struct agent *agent = adapter_get_agent(adapter);
 
 	if (device->agent)
 		agent_destroy(device->agent, FALSE);
 
+	if (agent && agent_is_busy(agent, device))
+		agent_cancel(agent);
+
 	g_slist_foreach(device->uuids, (GFunc) g_free, NULL);
 	g_slist_free(device->uuids);
 
@@ -124,12 +134,15 @@ static void device_free(gpointer user_data)
 static gboolean device_is_paired(struct btd_device *device)
 {
 	struct btd_adapter *adapter = device->adapter;
-	char filename[PATH_MAX + 1], *str;
+	char filename[PATH_MAX + 1], *str, addr[18];
 	gboolean ret;
-	const gchar *source = adapter_get_address(adapter);
+	bdaddr_t bdaddr;
+
+	adapter_get_address(adapter, &bdaddr);
+	ba2str(&bdaddr, addr);
 
 	create_name(filename, PATH_MAX, STORAGEDIR,
-			source, "linkkeys");
+			addr, "linkkeys");
 	str = textfile_caseget(filename, device->address);
 	ret = str ? TRUE : FALSE;
 	g_free(str);
@@ -146,9 +159,9 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	DBusMessageIter iter;
 	DBusMessageIter dict;
 	bdaddr_t src, dst;
-	char name[248];
+	char name[248], src_addr[18];
 	char **uuids;
-	const char *ptr, *source;
+	const char *ptr;
 	dbus_bool_t boolean;
 	uint32_t class;
 	int i;
@@ -173,15 +186,16 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	/* Name */
 	ptr = NULL;
 	memset(name, 0, sizeof(name));
-	source = adapter_get_address(adapter);
+	adapter_get_address(adapter, &src);
+	ba2str(&src, src_addr);
 
-	if (read_device_name(source, device->address, name) == 0) {
+	if (read_device_name(src_addr, device->address, name) == 0) {
 		ptr = name;
 		dbus_message_iter_append_dict_entry(&dict, "Name",
 				DBUS_TYPE_STRING, &ptr);
 	}
 
-	if (read_device_alias(source, device->address, name, sizeof(name)) > 0)
+	if (read_device_alias(src_addr, device->address, name, sizeof(name)) > 0)
 		ptr = name;
 
 	/* Alias: use Name if Alias doesn't exist */
@@ -189,7 +203,6 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		dbus_message_iter_append_dict_entry(&dict, "Alias",
 				DBUS_TYPE_STRING, &ptr);
 
-	str2ba(source, &src);
 	str2ba(device->address, &dst);
 
 	/* Class */
@@ -241,11 +254,15 @@ static DBusMessage *set_alias(DBusConnection *conn, DBusMessage *msg,
 {
 	struct btd_device *device = data;
 	struct btd_adapter *adapter = device->adapter;
-	const gchar *source = adapter_get_address(adapter);
+	char addr[18];
+	bdaddr_t bdaddr;
 	int err;
 
-	/* Remove alias if empty string */
-	err = write_device_alias(source, device->address,
+	adapter_get_address(adapter, &bdaddr);
+	ba2str(&bdaddr, addr);
+
+	 /* Remove alias if empty string */
+	err = write_device_alias(addr, device->address,
 			g_str_equal(alias, "") ? NULL : alias);
 	if (err < 0)
 		return g_dbus_create_error(msg,
@@ -264,9 +281,14 @@ static DBusMessage *set_trust(DBusConnection *conn, DBusMessage *msg,
 {
 	struct btd_device *device = data;
 	struct btd_adapter *adapter = device->adapter;
-	const gchar *source = adapter_get_address(adapter);
+	char addr[18];
+	bdaddr_t bdaddr;
 
-	write_trust(source, device->address, GLOBAL_TRUST, value);
+
+	adapter_get_address(adapter, &bdaddr);
+	ba2str(&bdaddr, addr);
+
+	write_trust(addr, device->address, GLOBAL_TRUST, value);
 
 	dbus_connection_emit_property_changed(conn, dbus_message_get_path(msg),
 					DEVICE_INTERFACE, "Trusted",
@@ -328,11 +350,11 @@ static void discover_services_req_exit(void *user_data)
 	struct btd_device *device = user_data;
 	struct btd_adapter *adapter = device->adapter;
 	bdaddr_t src, dst;
-	const gchar *source = adapter_get_address(adapter);
+
+	adapter_get_address(adapter, &src);
 
 	debug("DiscoverDevices requestor exited");
 
-	str2ba(source, &src);
 	str2ba(device->address, &dst);
 
 	bt_cancel_discovery(&src, &dst);
@@ -381,7 +403,8 @@ static DBusMessage *cancel_discover(DBusConnection *conn,
 	struct btd_device *device = user_data;
 	struct btd_adapter *adapter = device->adapter;
 	bdaddr_t src, dst;
-	const gchar *source = adapter_get_address(adapter);
+
+	adapter_get_address(adapter, &src);
 
 	if (!device->discov_active)
 		return g_dbus_create_error(msg,
@@ -395,7 +418,6 @@ static DBusMessage *cancel_discover(DBusConnection *conn,
 				ERROR_INTERFACE ".NotAuthorized",
 				"Not Authorized");
 
- 	str2ba(source, &src);
 	str2ba(device->address, &dst);
 
 	if (bt_cancel_discovery(&src, &dst) < 0)
@@ -621,9 +643,12 @@ void device_probe_drivers(struct btd_device *device, GSList *uuids, sdp_list_t *
 void device_remove_drivers(struct btd_device *device, GSList *uuids, sdp_list_t *recs)
 {
 	struct btd_adapter *adapter = device_get_adapter(device);
-	const gchar *src = adapter_get_address(adapter);
-	const gchar *dst = device_get_address(device);
 	GSList *list;
+	char src_addr[18];
+	bdaddr_t src;
+
+	adapter_get_address(adapter, &src);
+	ba2str(&src, src_addr);
 
 	debug("Remove drivers for %s", device->path);
 
@@ -649,7 +674,7 @@ void device_remove_drivers(struct btd_device *device, GSList *uuids, sdp_list_t
 			if (!rec)
 				continue;
 
-			delete_record(src, dst, rec->handle);
+			delete_record(src_addr, device->address, rec->handle);
 		}
 	}
 
@@ -779,9 +804,12 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 {
 	struct btd_device *device = req->device;
 	struct btd_adapter *adapter = device_get_adapter(device);
-	const gchar *src = adapter_get_address(adapter);
-	const gchar *dst = device_get_address(device);
 	sdp_list_t *seq;
+	char src_addr[18];
+	bdaddr_t src;
+
+	adapter_get_address(adapter, &src);
+	ba2str(&src, src_addr);
 
 	for (seq = recs; seq; seq = seq->next) {
 		sdp_record_t *rec = (sdp_record_t *) seq->data;
@@ -801,11 +829,23 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 		if (!uuid_str)
 			continue;
 
+		/* Driver uuid found */
+		l = g_slist_find_custom(req->uuids, uuid_str,
+				(GCompareFunc) strcasecmp);
+		if (l) {
+			char *uuid = l->data;
+
+			req->uuids = g_slist_remove(req->uuids, uuid);
+			g_free(uuid);
+		}
+
 		/* Check for duplicates */
 		if (sdp_list_find(req->records, rec, rec_cmp))
 			continue;
 
-		store_record(src, dst, rec);
+		debug("New record found: %s", uuid_str);
+
+		store_record(src_addr, device->address, rec);
 
 		/* Copy record */
 		if (sdp_gen_record_pdu(rec, &pdu) == 0) {
@@ -814,6 +854,7 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 
 			record = sdp_extract_pdu(pdu.data, pdu.data_size,
 						&scanned);
+			free(pdu.data);
 			req->records = sdp_list_append(req->records, record);
 		}
 
@@ -837,9 +878,8 @@ static void store(struct btd_device *device)
 	struct btd_adapter *adapter = device->adapter;
 	bdaddr_t src, dst;
 	char *str;
-	const gchar *source = adapter_get_address(adapter);
 
-	str2ba(source, &src);
+	adapter_get_address(adapter, &src);
 	str2ba(device->address, &dst);
 
 	if (!device->uuids) {
@@ -852,15 +892,11 @@ static void store(struct btd_device *device)
 	g_free(str);
 }
 
-static void browse_cb(sdp_list_t *recs, int err, gpointer user_data)
+static void search_cb(sdp_list_t *recs, int err, gpointer user_data)
 {
 	struct browse_req *req = user_data;
 	struct btd_device *device = req->device;
-	struct btd_adapter *adapter = device->adapter;
-	bdaddr_t src, dst;
-	uuid_t uuid;
 	DBusMessage *reply;
-	const gchar *source = adapter_get_address(adapter);
 
 	if (err < 0) {
 		error("%s: error updating services: %s (%d)",
@@ -870,22 +906,8 @@ static void browse_cb(sdp_list_t *recs, int err, gpointer user_data)
 
 	update_services(req, recs);
 
-	/* Public browsing successful or Single record requested */
-	if (req->browse == FALSE || (!req->search_uuid && recs))
-		goto probe;
-
-	if (uuid_list[++req->search_uuid]) {
-		sdp_uuid16_create(&uuid, uuid_list[req->search_uuid]);
-		str2ba(source, &src);
-		str2ba(device->address, &dst);
-		bt_search_service(&src, &dst, &uuid, browse_cb, user_data, NULL);
-		return;
-	}
-
-probe:
-
 	if (!req->uuids_added && !req->uuids_removed) {
-		debug("%s: No service found", device->path);
+		debug("%s: No service update", device->path);
 		goto proceed;
 	}
 
@@ -944,6 +966,78 @@ cleanup:
 	if (req->records)
 		sdp_list_free(req->records, (sdp_free_func_t) sdp_record_free);
 	g_free(req);
+
+}
+
+static void browse_cb(sdp_list_t *recs, int err, gpointer user_data)
+{
+	struct browse_req *req = user_data;
+	struct btd_device *device = req->device;
+	struct btd_adapter *adapter = device->adapter;
+	bdaddr_t src, dst;
+	uuid_t uuid;
+
+	/* Public browsing successful or Single record requested */
+	if (err < 0 || (!req->search_uuid && recs))
+		goto done;
+
+	update_services(req, recs);
+
+	adapter_get_address(adapter, &src);
+	str2ba(device->address, &dst);
+
+	/* Search for mandatory uuids */
+	if (uuid_list[++req->search_uuid]) {
+		sdp_uuid16_create(&uuid, uuid_list[req->search_uuid]);
+		bt_search_service(&src, &dst, &uuid, browse_cb, user_data, NULL);
+		return;
+	}
+
+	/* Search for drivers uuids */
+	if (req->uuids) {
+		char *uuid_str = req->uuids->data;
+
+		bt_string2uuid(&uuid, uuid_str);
+		req->uuids = g_slist_remove(req->uuids, uuid_str);
+		g_free(uuid_str);
+		bt_search_service(&src, &dst, &uuid, browse_cb, user_data, NULL);
+		return;
+	}
+
+	/* Search for l2cap uuid */
+	if (!req->records) {
+		sdp_uuid16_create(&uuid, L2CAP_UUID);
+		bt_search_service(&src, &dst, &uuid, search_cb, user_data, NULL);
+		return;
+	}
+
+done:
+	search_cb(recs, err, user_data);
+}
+
+static void init_browse(struct browse_req *req)
+{
+	GSList *l;
+
+	for (l = device_drivers; l; l = l->next) {
+		struct btd_device_driver *driver = l->data;
+		int i;
+
+		for (i = 0; driver->uuids[i]; i++) {
+			char *uuid;
+
+			if (g_slist_find_custom(req->uuids, driver->uuids[i],
+					(GCompareFunc) strcasecmp))
+				return;
+
+			uuid = g_strdup(driver->uuids[i]);
+			req->uuids = g_slist_append(req->uuids, uuid);
+		}
+	}
+
+	for (l = req->device->uuids; l; l = l->next)
+		req->uuids_removed = g_slist_append(req->uuids_removed,
+						l->data);
 }
 
 int device_browse(struct btd_device *device, DBusConnection *conn,
@@ -953,26 +1047,24 @@ int device_browse(struct btd_device *device, DBusConnection *conn,
 	struct browse_req *req;
 	bdaddr_t src, dst;
 	uuid_t uuid;
-	GSList *l;
-	const gchar *source = adapter_get_address(adapter);
+	bt_callback_t cb;
+
+	adapter_get_address(adapter, &src);
 
 	req = g_new0(struct browse_req, 1);
 	req->conn = dbus_connection_ref(conn);
 	req->msg = dbus_message_ref(msg);
 	req->device = device;
 
-	str2ba(source, &src);
 	str2ba(device->address, &dst);
 
 	if (search) {
 		memcpy(&uuid, search, sizeof(uuid_t));
-		req->browse = FALSE;
+		cb = search_cb;
 	} else {
 		sdp_uuid16_create(&uuid, uuid_list[req->search_uuid]);
-		req->browse = TRUE;
-		for (l = device->uuids; l; l = l->next)
-			req->uuids_removed = g_slist_append(req->uuids_removed,
-						l->data);
+		init_browse(req);
+		cb = browse_cb;
 	}
 
 	device->discov_active = 1;
@@ -995,12 +1087,9 @@ struct btd_adapter *device_get_adapter(struct btd_device *device)
 	return device->adapter;
 }
 
-const gchar *device_get_address(struct btd_device *device)
+void device_get_address(struct btd_device *device, bdaddr_t *bdaddr)
 {
-	if (!device)
-		return NULL;
-
-	return device->address;
+	str2ba(device->address, bdaddr);
 }
 
 const gchar *device_get_path(struct btd_device *device)
